from io import StringIO
import os
import re
import subprocess
from typing import Dict, List, Optional, Tuple
import configparser
import logging


from hid_firmware import HidMtAbstractFirmware

class HidMtDevice:
    def __init__(self, vendor_id: str, product_id: str, device_name: str):
        self.vendor_id = vendor_id
        self.product_id = product_id
        self.device_name = device_name

    def __str__(self) -> str:
        return f'{self.vendor_id}:{self.product_id} "{self.device_name}"'


class HidMtTools:
    @classmethod
    def detect_device(cls, device_name: str) -> Optional['HidMtAbstractFirmware']:
        props = cls.get_props_from_device_name(device_name)
        ctm = cls.__get_property_value(props, "Coordinate Transformation Matrix")
        lcm = cls.__get_property_value(props, "libinput Calibration Matrix")
        rot = cls.__get_property_value(props, "Rotation Angle")

        try:
            if rot is not None:
                rot = float(rot)
        except:
            rot = None

        if ctm is None:
            return None

        return HidMtAbstractFirmware(device_name, '', ctm, lcm, rot)
    
    @classmethod
    def get_props_from_device_name(cls, device_name: str) -> Dict[str, str]:
        output = subprocess.check_output(["xinput", "list-props", device_name]).decode()
        props = {}
        for line in output.split("\n"):
            if ":" in line:
                prop_name, prop_value = line.split(":", 1)
                prop_name = prop_name.strip()
                prop_value = prop_value.strip()
                props[prop_name] = prop_value
        return props

    @classmethod
    def __get_property_value(cls, props: Dict[str, str], name: str):
        for p in props:
            if name in p:
                return props[p]
        return None
    
    @classmethod
    def filter_hid_devices(cls, vendor_id: str) -> List[HidMtDevice]:
        hid_devices = []

        for f in os.listdir("/sys/class/hidraw"):
            file_path = "/sys/class/hidraw/" + f + "/device/uevent"
            if os.path.exists(file_path):
                content = open(file_path, "r").read()
                match = re.search(rf"HID_ID\s*=\s*0003:0000{vendor_id}:0000([0-9a-fA-F]{{4}})", content)
                driver = re.search(r"DRIVER=hid-multitouch", content)
                if match and driver:
                    product_id = match.group(1)
                    device_name = re.search(r"HID_NAME=(.+)", content)
                    if device_name:
                        hid_devices.append(HidMtDevice(vendor_id, product_id, device_name.group(1)))
        return hid_devices
    
    # call xinput and set the property
    @classmethod
    async def set_xinput_propery(cls, xinput_device_name: str, xinput_property_name: str, xinput_property_values):
        s = subprocess.run(["xinput", "set-prop", xinput_device_name, xinput_property_name, *xinput_property_values])
        
        # check return code and return status
        return_code = s.returncode
        return return_code == 0

    @classmethod
    def persist_settings(cls, hid_mt_device: 'HidMtAbstractFirmware'):
        
        pass

class HidConfigManager:
    DEBUG = False
    DEFAULT_ETC_CONFIG = '/etc/fortouch.conf'
    DEFAULT_UDEV_CONFIG = '/etc/udev/rules.d/99-hid-fortouch.rules'
    
    def __init__(self):
        self.file_path_etc = HidConfigManager.DEFAULT_ETC_CONFIG
        self.file_path_udev = HidConfigManager.DEFAULT_UDEV_CONFIG

    def __get_cfgwriter_path__(self):
        cfgwriter_path = None
        if os.path.exists("/usr/local/bin/cfgwriter"):
            cfgwriter_path = "/usr/local/bin/cfgwriter"
        elif os.path.exists("./exec/hid/cfgwriter"):
            cfgwriter_path = "./exec/hid/cfgwriter"
        elif self.DEBUG and os.path.exists("/tmp/cfgw_test"):
            cfgwriter_path = "/tmp/cfgw_test"
        return cfgwriter_path
    

    def load(self) -> dict:
        config = configparser.ConfigParser()
        config.read(self.file_path_etc)

        # Set default values
        default_values = {
            'device': None,
            'ct_matrix': '1, 0, 0, 0, 1, 0, 0, 0, 1',
            'screen_name': None
        }

        # Create a dictionary with default values and update it with the actual values from the config file
        result = default_values.copy()
        for key, value in default_values.items():
            if config.has_option("DEFAULT", key):
                cfg_value = config.get("DEFAULT", key, fallback=value)
                result[key] = cfg_value

        return result

    def save(self, device: Optional[str], matrix: str, screen_name: Optional[str]):
        config = configparser.ConfigParser()
        config["DEFAULT"] = {k: v for k, v in {
            "device": device,
            "ct_matrix": matrix,
            "screen_name": screen_name
        }.items() if v is not None}

        # 1. Generate the config file content as a string
        config_string = ""
        with StringIO() as f:
            config.write(f)
            config_string = f.getvalue()

        # 2. Check for the location of `cfgwriter`
        cfgwriter_path = self.__get_cfgwriter_path__()
        logging.info(f"cfgwriter was found at this location: {cfgwriter_path}")
        
        if not cfgwriter_path:
            raise FileNotFoundError("cfgwriter binary not found")

        # 3. Run `cfgwriter` and provide the string as its input
        process = subprocess.run([cfgwriter_path], input=config_string, text=True)
        process.check_returncode()

    def save_udev(self, firmware: HidMtAbstractFirmware):
        udev_rule = self.__get_udev_comment(firmware.device_name, firmware.vendor_id, firmware.product_id)
        udev_rule += self.__get_udev_rule(firmware.vendor_id, firmware.product_id, firmware.get_coord_matrix_udev_str())

        # 2. Check for the location of `cfgwriter`
        cfgwriter_path = self.__get_cfgwriter_path__()
        logging.info(f"cfgwriter was found at this location: {cfgwriter_path}")
        
        if not cfgwriter_path:
            raise FileNotFoundError("cfgwriter binary not found")

        # 3. Run `cfgwriter` and provide the string as its input
        process = subprocess.run([cfgwriter_path, '-u'], input=udev_rule, text=True)
        process.check_returncode()

    def load_udev(self, file_path: str) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        vendor_id = None
        product_id = None
        calibration_matrix = None

        try:
            with open(file_path, 'r') as file:
                rule = file.read()

                vendor_id_match = re.search(r'ENV{ID_VENDOR_ID}=="(.*?)"', rule)
                product_id_match = re.search(r'ENV{ID_MODEL_ID}=="(.*?)"', rule)
                calibration_matrix_match = re.search(r'ENV{LIBINPUT_CALIBRATION_MATRIX}="(.*?)"', rule)

                if vendor_id_match:
                    vendor_id = vendor_id_match.group(1)

                if product_id_match:
                    product_id = product_id_match.group(1)

                if calibration_matrix_match:
                    calibration_matrix = calibration_matrix_match.group(1)

        except FileNotFoundError:
            print(f"File '{file_path}' not found.")

        return vendor_id, product_id, calibration_matrix
    
    @classmethod
    def __get_udev_comment(cls, device_name: str, vendor_id: str, product_id: str,):
        return f'# ------- hid device: {device_name} | {vendor_id}:{product_id} -------\n'

    @classmethod
    def __get_udev_rule(cls, vendor_id: str, product_id: str, ct_matrix: str):
        # example: ACTION=="add|change", KERNEL=="event[0-9]*", ENV{ID_VENDOR_ID}=="0EEF", ENV{ID_MODEL_ID}=="C000", ENV{ID_INPUT_TOUCHSCREEN}="1", ENV{LIBINPUT_CALIBRATION_MATRIX}="-1 0 1 0 -1 1"
        return f'ACTION=="add|change", KERNEL=="event[0-9]*", ENV{{ID_VENDOR_ID}}=="{vendor_id}", ENV{{ID_MODEL_ID}}=="{product_id}", ENV{{ID_INPUT_TOUCHSCREEN}}="1", ENV{{LIBINPUT_CALIBRATION_MATRIX}}="{ct_matrix}"'