import asyncio
import os
import shlex
import stat
from traceback import print_exc
import fnmatch
import logging

from exec.tools import Platform, UnsupportedPlatform, uniplatform, InsufficientPermissions
from ilitek_fw_type_adapter import IlitekFirmwareTypeAdapter
from outparser import IliParser

TOOLS_PATH = ''
TOOL_NAME_LINUX = 'ilitek_ldv3_0_3_0_x86_x64'
TOOL_NAME_WINDOWS = 'Windows_Studio_ITS.exe'


def get_exe_path(current_platform: str) -> str:
    return os.path.join(os.getcwd(), 'exec/ili/' + current_platform + '/')

def get_exe_file_path(file_name: str, current_platform: str) -> str:
    return os.path.join(get_exe_path(current_platform), file_name)

def _check_setuid_set(exe_file) -> bool:
    try:
        return os.stat(exe_file).st_mode & stat.S_ISUID != 0
    except Exception as e:
        logging.error(f'Failed to check setuid on {exe_file}: {e}')
        return False

def _check_env_password() -> bool:
    try:
        return os.environ['FT_PWD'] != ''
    except Exception as e:
        logging.error(f'Failed to check env $FT_PWD: {e}')
        return False

async def run_linux_tool(exe_file, exe_path, exe_args):
    proc = None

    if _check_setuid_set(exe_file):
        logging.debug(f'RUN_ILITEK: setuid detected')
        proc = await asyncio.create_subprocess_exec(
            exe_path, *exe_args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=exe_path)
    elif _check_env_password():
        logging.debug(f'RUN_ILITEK: sudo with env password')
        cmd = f'printf "%s" "$FT_PWD" | sudo -S {exe_file} {" ".join(map(shlex.quote, exe_args))}'

        proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        cwd=exe_path)
    else:
        raise InsufficientPermissions("Linux tool requires setuid or env $FT_PWD")

    return proc

async def _run_windows_tool_bare(tool_path, exe_args):
    exe_file = TOOL_NAME_WINDOWS

    print(f"[ILI][WINDOWS] running tool {exe_file} located: {tool_path} with params {exe_args}")

    from subprocess import CREATE_NO_WINDOW # type: ignore
    proc = await asyncio.create_subprocess_exec(
        exe_file, *exe_args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        cwd=tool_path,
        creationflags=CREATE_NO_WINDOW)
    
    return proc

async def run_windows_tool(tool_path, *exe_args):
    exe_file = TOOL_NAME_WINDOWS
    exe_path = os.path.join(tool_path, exe_file)

    log_folder = os.path.join(tool_path, "Log")
    
    # check if "Log" folder exists, if not create it
    if not os.path.exists(log_folder):
        os.makedirs(log_folder)

    
    # delete all txt files in "Log" folder
    for file_name in os.listdir(log_folder):
        if file_name.endswith(".txt"):
            os.remove(os.path.join(log_folder, file_name))


    print(f"[ILI2][WINDOWS] running tool {exe_file} located: {tool_path} with params {exe_args}")
    try:
        #from subprocess import CREATE_NO_WINDOW
        proc = await asyncio.create_subprocess_exec(
            exe_path, *exe_args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=tool_path)
            #creationflags=CREATE_NO_WINDOW)

        await proc.communicate()
    except Exception as e:
        print(f'exception: {e}')
        print_exc()
    
    
    # check if there is a txt file in Log folder
    log_files = [file_name for file_name in os.listdir(log_folder) if file_name.endswith(".txt")]
    
    # if there is a txt file, read it into the variable
    if len(log_files) > 0:
        with open(os.path.join(log_folder, log_files[0]), "r") as f:
            log_contents = f.read()
        return log_contents
    
    # if there is no txt file, indicate error
    else:
        return None

async def run_tool(exe_name: str, *exe_args):
    current_platform = uniplatform()

    exe_path = get_exe_path(str(current_platform))
    exe_file = os.path.join(exe_path, exe_name)

    print(f"[ILI][{current_platform}] running tool {exe_name} located: {exe_path} with params {exe_args}")

    if current_platform == Platform.LINUX:

        return await run_linux_tool(exe_file, exe_path, exe_args)

    elif current_platform == Platform.WINDOWS:

        return await run_windows_tool(exe_path, *exe_args)

    else:
        raise UnsupportedPlatform
    
async def update_windows(fw_path: str):
    args = ['/hide', '/s', '/log', '/path=Log', fw_path]
    p = await run_tool(TOOL_NAME_WINDOWS, *args)
    if p is not None:
        if isinstance(p, str):
            ili_type = IliParser.parse_updatefw_windows(p)

    print('------- after ilitek update -------')

    return True

    
async def soft_reset_windows():
    ili_res = False
    args = ['/hide', '/sw_rst', '/log', '/path=Log']
    p = await run_tool(TOOL_NAME_WINDOWS, *args)
    if p is not None:
        if isinstance(p, str):
            ili_res = IliParser.parse_softreset_windows(p)

    print(f'------- after ilitek soft reset {ili_res} -------')

    return ili_res

async def identify_windows():
    args = ['/hide', '/i', '/log', '/path=Log']
    ili_type = None
    p = await run_tool(TOOL_NAME_WINDOWS, *args)

    if p is not None:
        if isinstance(p, str):
            ili_type = IliParser.parse_fw_type_windows(p)

    return IlitekFirmwareTypeAdapter(ili_type)


async def identify_linux():
    args_v3 = ['PanelInfor', 'USB', 'V3', 'null', 'null']
    args_v6 = ['PanelInfor', 'USB', 'V6', 'null', 'null']

    fwtype = None
    print('ident')

    p = await run_tool(TOOL_NAME_LINUX, *args_v3)

    if p is not None and p.stdout: # type: ignore
        value = await p.stdout.read() # type: ignore
        fwtype = IliParser.parse_fw_type_linux(value)
        print(f'OUTPUT: {value}')

    return IlitekFirmwareTypeAdapter(fwtype)

"""
Return hex file in ParaData directory.
Rules are:
1. If there are no files with the "hex" extension, return None.
2. If there is only one file with the "hex" extension, return that file's full path.
3. If there are multiple files with the "hex" extension, filter them based on the "version" parameter.
4. If there is only one file with the "version" parameter, return that file's full path.
5. If there are multiple files with the "hex" extension but none with the "version" parameter, return the full path of the first file in the list.
"""
def find_hex_file(path: str, version=None):
    # get a list of all files in the directory
    files = []
    try:
        files = os.listdir(path)
    except:
        logging.error(f'Error when listing files in "{path}"')
        return None
    
    # filter files based on extension
    hex_files = fnmatch.filter(files, '*.hex')
    
    # return None if no hex files found
    if not hex_files:
        return None
    
    # return the only hex file found
    if len(hex_files) == 1:
        return os.path.join(path, hex_files[0])
    
    # filter files based on version parameter
    version_files = fnmatch.filter(hex_files, f'*{version}*')
    
    # return the only version file found
    if len(version_files) == 1:
        return os.path.join(path, version_files[0])
    
    # return the first file in the list if no version files found
    return os.path.join(path, hex_files[0])

def get_hex_file_directory(serie: str, version: str):
    return os.path.join('exec', 'ili', 'windows', 'ParaData', serie, f'v{version}')