from enum import Enum
import platform
import sys
import asyncio
import os
import tempfile
import logging

class UnsupportedPlatform(Exception):
    pass

class InsufficientPermissions(Exception):
    pass

def create_temp() -> str:
    try:
        return tempfile.mkdtemp(suffix='fortouch', prefix='wrk')
    except Exception:
        return ''

class Platform(Enum):
    LINUX = 0
    WINDOWS = 1
    MAC = 2
    UNKNOWN = 3

    def __str__(self):
        if self == Platform.LINUX:
            return "linux"
        if self == Platform.WINDOWS:
            return "windows"
        if self == Platform.MAC:
            return "mac"
        return "unknown"

def uniplatform() -> Platform:
    if "linux" in sys.platform:
        return Platform.LINUX
    elif "win" in sys.platform:
        return Platform.WINDOWS
    elif "darwin" in sys.platform:
        return Platform.MAC
    else:
        return Platform.UNKNOWN
    
def uses_wayland() -> bool:
    is_wayland = os.environ.get('WAYLAND_DISPLAY') is not None
    return uniplatform() == Platform.LINUX and is_wayland


class Architecture(Enum):
    BIT_32 = 0
    BIT_64 = 1
    UNKNOWN = 2

    def __str__(self):
        if self == Architecture.BIT_32:
            return "32bit"
        if self == Architecture.BIT_64:
            return "64bit"
        return "unknown"

def architecture() -> Architecture:
    # taken from: https://docs.python.org/3/library/platform.html#platform.architecture
    # TODO: test and fix if necessary
    arch = platform.architecture()[0]
    if "32bit" in arch:
        return Architecture.BIT_32
    if "64bit" in arch:
        return Architecture.BIT_64
    return Architecture.UNKNOWN


def get_exe_path(current_platform: str) -> str:
    return os.path.join(os.getcwd(), './exec/' + current_platform + '/')

def get_exe_file_path(file_name: str, current_platform: str) -> str:
    return os.path.join(get_exe_path(current_platform), file_name)


def tool_exists(exe_name: str):
    current_platform = uniplatform()
    exe_path = get_exe_path(str(current_platform))
    exe_file = os.path.join(exe_path, exe_name)
    
    return os.path.exists(exe_file) or os.path.exists(exe_file + ".exe")


async def run_tool(exe_name: str, *exe_args):
    current_platform = uniplatform()

    exe_path = get_exe_path(str(current_platform))
    exe_file = os.path.join(exe_path, exe_name)
    cwd = os.path.dirname(exe_file)

    logging.getLogger("SiS tools").info(f"[{current_platform}] running tool {exe_name} located: {exe_path} with params {exe_args}")

    if current_platform == Platform.LINUX:
        
        exe_env = {**os.environ}
        rest_ld = ''
        if 'LD_LIBRARY_PATH' in exe_env:
            rest_ld = ':' + exe_env['LD_LIBRARY_PATH']
            
        exe_env['LD_LIBRARY_PATH'] = os.path.join(os.getcwd(), './exec/linux') + rest_ld

        proc = await asyncio.create_subprocess_exec(
            exe_file, *exe_args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            limit=512,
            env=exe_env,
            cwd=cwd)

        return proc

    elif current_platform == Platform.WINDOWS:
        
        from subprocess import CREATE_NO_WINDOW # type: ignore
        proc = await asyncio.create_subprocess_exec(
            exe_file, *exe_args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=cwd,
            creationflags=CREATE_NO_WINDOW)
        
        return proc

    else:
        raise UnsupportedPlatform


def parse_int_or_hex(input_str: str) -> int:
    if input_str.lower().startswith('0x'):
        number = int(input_str, 16)
    else:
        number = int(input_str)
    return number