import os
import platform
from enum import Enum
import subprocess

class XSessionType(Enum):
    OS_INVALID = 0
    X_OK = 1
    X_NO_XINPUT = 2
    WAYLAND = 3

class XSessionDetector:
    @classmethod
    def detect(cls):
        if platform.system() != "Linux":
            return XSessionType.OS_INVALID
        wayland_status = cls.__detect_wayland()
        if wayland_status == XSessionType.WAYLAND:
            return wayland_status
        else:
            xinput_status = cls.__detect_xinput()
            if xinput_status == XSessionType.X_OK:
                return xinput_status
            else:
                return XSessionType.X_NO_XINPUT
            

    @staticmethod
    def __detect_xinput(verbose:bool=False):
        pstdout = None
        pstderr = None
        
        if verbose:
            pstdout = subprocess.PIPE
            pstderr = subprocess.STDOUT

        try:
            retcode = subprocess.call(['xinput', '--version'], stdout=pstdout, stderr=pstderr)
            if retcode == 0:
                return XSessionType.X_OK
            else:
                return XSessionType.X_NO_XINPUT
        except OSError:
            return XSessionType.X_NO_XINPUT

    @staticmethod
    def __detect_wayland():
        if "WAYLAND_DISPLAY" in os.environ or os.getenv("XDG_SESSION_TYPE") == "wayland":
            return XSessionType.WAYLAND
        else:
            return XSessionType.X_NO_XINPUT


def main():
    session_type = XSessionDetector.detect()
    if session_type == XSessionType.OS_INVALID:
        print("This program only works on Linux.")
    elif session_type == XSessionType.X_OK:
        print("Running on Xorg.")
    elif session_type == XSessionType.X_NO_XINPUT:
        print("Running on Xorg without xinput.")
    elif session_type == XSessionType.WAYLAND:
        print("Running on Wayland.")


if __name__ == "__main__":
    main()
